/**
 * Copyright (c) 2017-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */
package com.ebmwebsourcing.easycommons.stream;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;

/**
 * This is a non-synchronized version of {@link ByteArrayOutputStream}.
 *
 * It also gives access to the underlying byte array.
 *
 * @author Victor Noel
 */
public class EasyByteArrayOutputStream extends ByteArrayOutputStream {

    public EasyByteArrayOutputStream() {
        super();
    }

    public EasyByteArrayOutputStream(int size) {
        super(size);
    }

    public byte[] getBuf() {
        return buf;
    }

    public int getCount() {
        return count;
    }

    private void ensureCapacity(int minCapacity) {
        // overflow-conscious code
        if (minCapacity - buf.length > 0)
            grow(minCapacity);
    }

    private static final int MAX_ARRAY_SIZE = Integer.MAX_VALUE - 8;

    private void grow(int minCapacity) {
        // overflow-conscious code
        int oldCapacity = buf.length;
        int newCapacity = oldCapacity << 1;
        if (newCapacity - minCapacity < 0)
            newCapacity = minCapacity;
        if (newCapacity - MAX_ARRAY_SIZE > 0)
            newCapacity = hugeCapacity(minCapacity);
        buf = Arrays.copyOf(buf, newCapacity);
    }

    private static int hugeCapacity(int minCapacity) {
        if (minCapacity < 0) // overflow
            throw new OutOfMemoryError();
        return (minCapacity > MAX_ARRAY_SIZE) ? Integer.MAX_VALUE : MAX_ARRAY_SIZE;
    }

    @Override
    @SuppressWarnings("sync-override")
    public void write(int b) {
        ensureCapacity(count + 1);
        buf[count] = (byte) b;
        count += 1;
    }

    @Override
    @SuppressWarnings("sync-override")
    public void write(byte b[], int off, int len) {
        if ((off < 0) || (off > b.length) || (len < 0) || ((off + len) - b.length > 0)) {
            throw new IndexOutOfBoundsException();
        }
        ensureCapacity(count + len);
        System.arraycopy(b, off, buf, count, len);
        count += len;
    }

    @Override
    @SuppressWarnings("sync-override")
    public void writeTo(OutputStream out) throws IOException {
        out.write(buf, 0, count);
    }

    @Override
    @SuppressWarnings("sync-override")
    public void reset() {
        count = 0;
    }

    @Override
    @SuppressWarnings("sync-override")
    public byte[] toByteArray() {
        return Arrays.copyOf(buf, count);
    }

    public byte[] toRawByteArray() {
        if (buf.length == count) {
            return buf;
        } else {
            return toByteArray();
        }
    }

    public ByteArrayInputStream toByteArrayInputStream() {
        return new ByteArrayInputStream(buf, 0, count);
    }

    @Override
    @SuppressWarnings("sync-override")
    public int size() {
        return count;
    }

    @Override
    @SuppressWarnings("sync-override")
    public String toString() {
        return new String(buf, 0, count);
    }

    @Override
    @SuppressWarnings("sync-override")
    public String toString(String charsetName) throws UnsupportedEncodingException {
        return new String(buf, 0, count, charsetName);
    }

    @Deprecated
    @Override
    @SuppressWarnings("sync-override")
    public String toString(int hibyte) {
        return new String(buf, hibyte, 0, count);
    }

    @Override
    public void close() {
        // nothing to do
    }
}
