/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */
package com.ebmwebsourcing.easycommons.stream;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringWriter;
import java.util.Arrays;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * @author ofabre
 * 
 */
public class StreamHelper {

    private static final int EOF = -1;

    /**
     * Default value is 2048.
     */
    public static final int DEFAULT_BUFFER_SIZE = 2048;

    /**
     * Copies information from the input stream to the output stream using a
     * default buffer size of 2048 bytes.
     * 
     * @param fromStream
     *            the {@link InputStream} to read
     * @param toStream
     *            the {@link OutputStream} to write
     */
    public static void copy(InputStream fromStream, OutputStream toStream) {
        copy(fromStream, toStream, DEFAULT_BUFFER_SIZE);
    }

    /**
     * Copies information from the input stream to the output stream using the
     * specified buffer size
     * 
     * @param fromStream
     *            the {@link InputStream} to read
     * @param toStream
     *            the {@link OutputStream} to write
     * @param bufferSize
     *            size of read blocks
     */
    public static void copy(InputStream fromStream, OutputStream toStream, int bufferSize) {
        assert fromStream != null;
        assert toStream != null;
        assert bufferSize > 0;

        int n = 0;
        byte[] buf = new byte[bufferSize];
        try {
            while (EOF != (n = fromStream.read(buf))) {
                toStream.write(buf, 0, n);
            }
            toStream.flush();
        } catch (IOException e) {
            throw new UncheckedException("Cannot read or write bytes on given streams", e);
        }
    }

    /**
     * @return a byte[] containing the information contained in the specified
     *          InputStream.
     */
    public static byte[] getBytes(InputStream input) {
        int length;
        // Note we don't call available() on any InputStream because of
        // buggy implementations such as our own ReaderInputStream...
        if (input instanceof ByteArrayInputStream) {
            length = ((ByteArrayInputStream) input).available();
        } else {
            // default used by ByteArrayOutputStream
            length = 32;
        }

        try (final EasyByteArrayOutputStream result = new EasyByteArrayOutputStream(length)) {
            copy(input, result);
            return result.toRawByteArray();
        }
    }

    /**
     * Contrary to {@link #getBytes(InputStream)}, this does not read the stream where it was at but all the bytes
     * inside!
     * 
     * @throws IllegalStateException
     *             if the implementation of {@link ByteArrayInputStream} is unexpected
     */
    public static byte[] getAllBytes(ByteArrayInputStream input) {
        try {
            // these two fields are immutable in a ByteArrayInputStream
            byte[] buf = (byte[]) ReflectionHelper.getFieldValue(ByteArrayInputStream.class, input, "buf", false);
            int count = (int) ReflectionHelper.getFieldValue(ByteArrayInputStream.class, input, "count", false);

            return Arrays.copyOf(buf, count);
        } catch (final Exception e) {
            throw new IllegalStateException("Unexpected implementation of ByteArrayInputStream!", e);
        }
    }

    /**
     * Return the {@link String} representation of an {@link InputStream}
     * <p>
     * NOTE: At the end, the {@link InputStream} is closed.
     * </p>
     * 
     * @param is
     * @return Return the {@link String} representation of an
     *         {@link InputStream}
     * @throws IOException
     */
    public static String getString(InputStream is) throws IOException {
        StringWriter w = null;
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(is));
            w = new StringWriter();
            char[] buffer = new char[1024];
            int n;
            while ((n = reader.read(buffer)) != -1) {
                w.write(buffer, 0, n);
            }
        } finally {
            is.close();
        }
        return w.toString();
    }
}
